@@ -1,8 +1,8 @@
 import { assert } from './assert';
-import { Messages} from './messages';
-import { Character} from './character';
-import { Token } from './token';
+import { Character } from './character';
 import { ErrorHandler } from './error-handler';
+import { Messages } from './messages';
+import { Token } from './token';
 
 function hexValue(ch: string): number {
     return '0123456789abcdef'.indexOf(ch.toLowerCase());
@@ -12,24 +12,58 @@ function octalValue(ch: string): number {
     return '01234567'.indexOf(ch);
 }
 
+export interface Position {
+    line: number;
+    column: number;
+}
+
+export interface SourceLocation {
+    start: Position;
+    end: Position;
+    source?: string;
+}
+
 export interface Comment {
     multiLine: boolean;
     slice: number[];
-    range: number[];
-    loc: any;
+    range: [number, number];
+    loc: SourceLocation;
+}
+
+export interface RawToken {
+    type: Token;
+    value: string | number;
+    pattern?: string;
+    flags?: string;
+    regex?: RegExp | null;
+    octal?: boolean;
+    cooked?: string;
+    head?: boolean;
+    tail?: boolean;
+    lineNumber: number;
+    lineStart: number;
+    start: number;
+    end: number;
+}
+
+interface ScannerState {
+    index: number;
+    lineNumber: number;
+    lineStart: number;
 }
 
 export class Scanner {
 
-    source: string;
-    errorHandler: ErrorHandler;
+    readonly source: string;
+    readonly errorHandler: ErrorHandler;
     trackComment: boolean;
 
-    length: number;
     index: number;
     lineNumber: number;
     lineStart: number;
-    curlyStack: string[];
+    private curlyStack: string[];
+
+    private readonly length: number;
 
     constructor(code: string, handler: ErrorHandler) {
         this.source = code;
@@ -41,26 +75,40 @@ export class Scanner {
         this.lineNumber = (code.length > 0) ? 1 : 0;
         this.lineStart = 0;
         this.curlyStack = [];
-    };
+    }
 
-    eof(): boolean {
+    public saveState(): ScannerState {
+        return {
+            index: this.index,
+            lineNumber: this.lineNumber,
+            lineStart: this.lineStart
+        };
+    }
+
+    public restoreState(state: ScannerState): void {
+        this.index = state.index;
+        this.lineNumber = state.lineNumber;
+        this.lineStart = state.lineStart;
+    }
+
+    public eof(): boolean {
         return this.index >= this.length;
-    };
+    }
 
-    throwUnexpectedToken(message = Messages.UnexpectedTokenIllegal) {
-        this.errorHandler.throwError(this.index, this.lineNumber,
+    public throwUnexpectedToken(message = Messages.UnexpectedTokenIllegal): never {
+        return this.errorHandler.throwError(this.index, this.lineNumber,
             this.index - this.lineStart + 1, message);
-    };
+    }
 
-    tolerateUnexpectedToken() {
+    private tolerateUnexpectedToken(message = Messages.UnexpectedTokenIllegal) {
         this.errorHandler.tolerateError(this.index, this.lineNumber,
-            this.index - this.lineStart + 1, Messages.UnexpectedTokenIllegal);
-    };
+            this.index - this.lineStart + 1, message);
+    }
 
-    // ECMA-262 11.4 Comments
+    // https://tc39.github.io/ecma262/#sec-comments
 
-    skipSingleLineComment(offset: number): Comment[] {
-        let comments: Comment[];
+    private skipSingleLineComment(offset: number): Comment[] {
+        let comments: Comment[] = [];
         let start, loc;
 
         if (this.trackComment) {
@@ -76,7 +124,7 @@ export class Scanner {
         }
 
         while (!this.eof()) {
-            let ch = this.source.charCodeAt(this.index);
+            const ch = this.source.charCodeAt(this.index);
             ++this.index;
             if (Character.isLineTerminator(ch)) {
                 if (this.trackComment) {
@@ -116,10 +164,10 @@ export class Scanner {
         }
 
         return comments;
-    };
+    }
 
-    skipMultiLineComment(): Comment[] {
-        let comments: Comment[];
+    private skipMultiLineComment(): Comment[] {
+        let comments: Comment[] = [];
         let start, loc;
 
         if (this.trackComment) {
@@ -185,9 +233,9 @@ export class Scanner {
 
         this.tolerateUnexpectedToken();
         return comments;
-    };
+    }
 
-    scanComments() {
+    public scanComments() {
         let comments;
         if (this.trackComment) {
             comments = [];
@@ -253,11 +301,11 @@ export class Scanner {
         }
 
         return comments;
-    };
+    }
 
-    // ECMA-262 11.6.2.2 Future Reserved Words
+    // https://tc39.github.io/ecma262/#sec-future-reserved-words
 
-    isFutureReservedWord(id: string): boolean {
+    public isFutureReservedWord(id: string): boolean {
         switch (id) {
             case 'enum':
             case 'export':
@@ -267,9 +315,9 @@ export class Scanner {
             default:
                 return false;
         }
-    };
+    }
 
-    isStrictModeReservedWord(id: string): boolean {
+    public isStrictModeReservedWord(id: string): boolean {
         switch (id) {
             case 'implements':
             case 'interface':
@@ -284,15 +332,15 @@ export class Scanner {
             default:
                 return false;
         }
-    };
+    }
 
-    isRestrictedWord(id: string): boolean {
+    public isRestrictedWord(id: string): boolean {
         return id === 'eval' || id === 'arguments';
-    };
+    }
 
-    // ECMA-262 11.6.2.1 Keywords
+    // https://tc39.github.io/ecma262/#sec-keywords
 
-    isKeyword(id: string): boolean {
+    private isKeyword(id: string): boolean {
         switch (id.length) {
             case 2:
                 return (id === 'if') || (id === 'in') || (id === 'do');
@@ -318,23 +366,23 @@ export class Scanner {
             default:
                 return false;
         }
-    };
+    }
 
-    codePointAt(i: number): number {
+    private codePointAt(i: number): number {
         let cp = this.source.charCodeAt(i);
 
         if (cp >= 0xD800 && cp <= 0xDBFF) {
-            let second = this.source.charCodeAt(i + 1);
+            const second = this.source.charCodeAt(i + 1);
             if (second >= 0xDC00 && second <= 0xDFFF) {
-                let first = cp;
+                const first = cp;
                 cp = (first - 0xD800) * 0x400 + second - 0xDC00 + 0x10000;
             }
         }
 
         return cp;
-    };
+    }
 
-    scanHexEscape(prefix: string): string {
+    private scanHexEscape(prefix: string): string | null {
         const len = (prefix === 'u') ? 4 : 2;
         let code = 0;
 
@@ -342,13 +390,13 @@ export class Scanner {
             if (!this.eof() && Character.isHexDigit(this.source.charCodeAt(this.index))) {
                 code = code * 16 + hexValue(this.source[this.index++]);
             } else {
-                return '';
+                return null;
             }
         }
         return String.fromCharCode(code);
-    };
+    }
 
-    scanUnicodeCodePointEscape(): string {
+    private scanUnicodeCodePointEscape(): string {
         let ch = this.source[this.index];
         let code = 0;
 
@@ -370,9 +418,9 @@ export class Scanner {
         }
 
         return Character.fromCodePoint(code);
-    };
+    }
 
-    getIdentifier(): string {
+    private getIdentifier(): string {
         const start = this.index++;
         while (!this.eof()) {
             const ch = this.source.charCodeAt(this.index);
@@ -393,9 +441,9 @@ export class Scanner {
         }
 
         return this.source.slice(start, this.index);
-    };
+    }
 
-    getComplexIdentifier(): string {
+    private getComplexIdentifier(): string {
         let cp = this.codePointAt(this.index);
         let id = Character.fromCodePoint(cp);
         this.index += id.length;
@@ -412,8 +460,7 @@ export class Scanner {
                 ch = this.scanUnicodeCodePointEscape();
             } else {
                 ch = this.scanHexEscape('u');
-                cp = ch.charCodeAt(0);
-                if (!ch || ch === '\\' || !Character.isIdentifierStart(cp)) {
+                if (ch === null || ch === '\\' || !Character.isIdentifierStart(ch.charCodeAt(0))) {
                     this.throwUnexpectedToken();
                 }
             }
@@ -441,8 +488,7 @@ export class Scanner {
                     ch = this.scanUnicodeCodePointEscape();
                 } else {
                     ch = this.scanHexEscape('u');
-                    cp = ch.charCodeAt(0);
-                    if (!ch || ch === '\\' || !Character.isIdentifierPart(cp)) {
+                    if (ch === null || ch === '\\' || !Character.isIdentifierPart(ch.charCodeAt(0))) {
                         this.throwUnexpectedToken();
                     }
                 }
@@ -451,9 +497,9 @@ export class Scanner {
         }
 
         return id;
-    };
+    }
 
-    octalToDecimal(ch: string) {
+    private octalToDecimal(ch: string) {
         // \0 is not octal escape sequence
         let octal = (ch !== '0');
         let code = octalValue(ch);
@@ -473,11 +519,11 @@ export class Scanner {
             code: code,
             octal: octal
         };
-    };
+    }
 
-    // ECMA-262 11.6 Names and Keywords
+    // https://tc39.github.io/ecma262/#sec-names-and-keywords
 
-    scanIdentifier() {
+    private scanIdentifier(): RawToken {
         let type: Token;
         const start = this.index;
 
@@ -498,6 +544,13 @@ export class Scanner {
             type = Token.Identifier;
         }
 
+        if (type !== Token.Identifier && (start + id.length !== this.index)) {
+            const restore = this.index;
+            this.index = start;
+            this.tolerateUnexpectedToken(Messages.InvalidEscapedReservedWord);
+            this.index = restore;
+        }
+
         return {
             type: type,
             value: id,
@@ -506,19 +559,12 @@ export class Scanner {
             start: start,
             end: this.index
         };
-    };
+    }
 
-    // ECMA-262 11.7 Punctuators
+    // https://tc39.github.io/ecma262/#sec-punctuators
 
-    scanPunctuator() {
-        const token = {
-            type: Token.Punctuator,
-            value: '',
-            lineNumber: this.lineNumber,
-            lineStart: this.lineStart,
-            start: this.index,
-            end: this.index
-        };
+    private scanPunctuator(): RawToken {
+        const start = this.index;
 
         // Check for most common single-character punctuators.
         let str = this.source[this.index];
@@ -590,28 +636,33 @@ export class Scanner {
                 }
         }
 
-        if (this.index === token.start) {
+        if (this.index === start) {
             this.throwUnexpectedToken();
         }
 
-        token.end = this.index;
-        token.value = str;
-        return token;
-    };
+        return {
+            type: Token.Punctuator,
+            value: str,
+            lineNumber: this.lineNumber,
+            lineStart: this.lineStart,
+            start: start,
+            end: this.index
+        };
+    }
 
-    // ECMA-262 11.8.3 Numeric Literals
+    // https://tc39.github.io/ecma262/#sec-literals-numeric-literals
 
-    scanHexLiteral(start: number) {
-        let number = '';
+    private scanHexLiteral(start: number): RawToken {
+        let num = '';
 
         while (!this.eof()) {
             if (!Character.isHexDigit(this.source.charCodeAt(this.index))) {
                 break;
             }
-            number += this.source[this.index++];
+            num += this.source[this.index++];
         }
 
-        if (number.length === 0) {
+        if (num.length === 0) {
             this.throwUnexpectedToken();
         }
 
@@ -621,16 +672,16 @@ export class Scanner {
 
         return {
             type: Token.NumericLiteral,
-            value: parseInt('0x' + number, 16),
+            value: parseInt('0x' + num, 16),
             lineNumber: this.lineNumber,
             lineStart: this.lineStart,
             start: start,
             end: this.index
         };
-    };
+    }
 
-    scanBinaryLiteral(start: number) {
-        let number = '';
+    private scanBinaryLiteral(start: number): RawToken {
+        let num = '';
         let ch;
 
         while (!this.eof()) {
@@ -638,10 +689,10 @@ export class Scanner {
             if (ch !== '0' && ch !== '1') {
                 break;
             }
-            number += this.source[this.index++];
+            num += this.source[this.index++];
         }
 
-        if (number.length === 0) {
+        if (num.length === 0) {
             // only 0b or 0B
             this.throwUnexpectedToken();
         }
@@ -656,21 +707,21 @@ export class Scanner {
 
         return {
             type: Token.NumericLiteral,
-            value: parseInt(number, 2),
+            value: parseInt(num, 2),
             lineNumber: this.lineNumber,
             lineStart: this.lineStart,
             start: start,
             end: this.index
         };
-    };
+    }
 
-    scanOctalLiteral(prefix: string, start: number) {
-        let number = '';
+    private scanOctalLiteral(prefix: string, start: number): RawToken {
+        let num = '';
         let octal = false;
 
         if (Character.isOctalDigit(prefix.charCodeAt(0))) {
             octal = true;
-            number = '0' + this.source[this.index++];
+            num = '0' + this.source[this.index++];
         } else {
             ++this.index;
         }
@@ -679,10 +730,10 @@ export class Scanner {
             if (!Character.isOctalDigit(this.source.charCodeAt(this.index))) {
                 break;
             }
-            number += this.source[this.index++];
+            num += this.source[this.index++];
         }
 
-        if (!octal && number.length === 0) {
+        if (!octal && num.length === 0) {
             // only 0o or 0O
             this.throwUnexpectedToken();
         }
@@ -693,16 +744,16 @@ export class Scanner {
 
         return {
             type: Token.NumericLiteral,
-            value: parseInt(number, 8),
+            value: parseInt(num, 8),
             octal: octal,
             lineNumber: this.lineNumber,
             lineStart: this.lineStart,
             start: start,
             end: this.index
         };
-    };
+    }
 
-    isImplicitOctalLiteral(): boolean {
+    private isImplicitOctalLiteral(): boolean {
         // Implicit octal, unless there is a non-octal digit.
         // (Annex B.1.1 on Numeric Literals)
         for (let i = this.index + 1; i < this.length; ++i) {
@@ -716,24 +767,24 @@ export class Scanner {
         }
 
         return true;
-    };
+    }
 
-    scanNumericLiteral() {
+    private scanNumericLiteral(): RawToken {
         const start = this.index;
         let ch = this.source[start];
         assert(Character.isDecimalDigit(ch.charCodeAt(0)) || (ch === '.'),
             'Numeric literal must start with a decimal digit or a decimal point');
 
-        let number = '';
+        let num = '';
         if (ch !== '.') {
-            number = this.source[this.index++];
+            num = this.source[this.index++];
             ch = this.source[this.index];
 
             // Hex number starts with '0x'.
             // Octal number starts with '0'.
             // Octal number in ES6 starts with '0o'.
             // Binary number in ES6 starts with '0b'.
-            if (number === '0') {
+            if (num === '0') {
                 if (ch === 'x' || ch === 'X') {
                     ++this.index;
                     return this.scanHexLiteral(start);
@@ -754,29 +805,29 @@ export class Scanner {
             }
 
             while (Character.isDecimalDigit(this.source.charCodeAt(this.index))) {
-                number += this.source[this.index++];
+                num += this.source[this.index++];
             }
             ch = this.source[this.index];
         }
 
         if (ch === '.') {
-            number += this.source[this.index++];
+            num += this.source[this.index++];
             while (Character.isDecimalDigit(this.source.charCodeAt(this.index))) {
-                number += this.source[this.index++];
+                num += this.source[this.index++];
             }
             ch = this.source[this.index];
         }
 
         if (ch === 'e' || ch === 'E') {
-            number += this.source[this.index++];
+            num += this.source[this.index++];
 
             ch = this.source[this.index];
             if (ch === '+' || ch === '-') {
-                number += this.source[this.index++];
+                num += this.source[this.index++];
             }
             if (Character.isDecimalDigit(this.source.charCodeAt(this.index))) {
                 while (Character.isDecimalDigit(this.source.charCodeAt(this.index))) {
-                    number += this.source[this.index++];
+                    num += this.source[this.index++];
                 }
             } else {
                 this.throwUnexpectedToken();
@@ -789,17 +840,17 @@ export class Scanner {
 
         return {
             type: Token.NumericLiteral,
-            value: parseFloat(number),
+            value: parseFloat(num),
             lineNumber: this.lineNumber,
             lineStart: this.lineStart,
             start: start,
             end: this.index
         };
-    };
+    }
 
-    // ECMA-262 11.8.4 String Literals
+    // https://tc39.github.io/ecma262/#sec-literals-string-literals
 
-    scanStringLiteral() {
+    private scanStringLiteral(): RawToken {
         const start = this.index;
         let quote = this.source[start];
         assert((quote === '\'' || quote === '"'),
@@ -820,18 +871,24 @@ export class Scanner {
                 if (!ch || !Character.isLineTerminator(ch.charCodeAt(0))) {
                     switch (ch) {
                         case 'u':
-                        case 'x':
                             if (this.source[this.index] === '{') {
                                 ++this.index;
                                 str += this.scanUnicodeCodePointEscape();
                             } else {
                                 const unescaped = this.scanHexEscape(ch);
-                                if (!unescaped) {
+                                if (unescaped === null) {
                                     this.throwUnexpectedToken();
                                 }
                                 str += unescaped;
                             }
                             break;
+                        case 'x':
+                            const unescaped = this.scanHexEscape(ch);
+                            if (unescaped === null) {
+                                this.throwUnexpectedToken(Messages.InvalidHexEscapeSequence);
+                            }
+                            str += unescaped;
+                            break;
                         case 'n':
                             str += '\n';
                             break;
@@ -895,16 +952,16 @@ export class Scanner {
             start: start,
             end: this.index
         };
-    };
+    }
 
-    // ECMA-262 11.8.6 Template Literal Lexical Components
+    // https://tc39.github.io/ecma262/#sec-template-literal-lexical-components
 
-    scanTemplate() {
+    private scanTemplate(): RawToken {
         let cooked = '';
         let terminated = false;
-        let start = this.index;
+        const start = this.index;
 
-        let head = (this.source[start] === '`');
+        const head = (this.source[start] === '`');
         let tail = false;
         let rawOffset = 2;
 
@@ -939,14 +996,13 @@ export class Scanner {
                             cooked += '\t';
                             break;
                         case 'u':
-                        case 'x':
                             if (this.source[this.index] === '{') {
                                 ++this.index;
                                 cooked += this.scanUnicodeCodePointEscape();
                             } else {
                                 const restore = this.index;
                                 const unescaped = this.scanHexEscape(ch);
-                                if (unescaped) {
+                                if (unescaped !== null) {
                                     cooked += unescaped;
                                 } else {
                                     this.index = restore;
@@ -954,6 +1010,13 @@ export class Scanner {
                                 }
                             }
                             break;
+                        case 'x':
+                            const unescaped = this.scanHexEscape(ch);
+                            if (unescaped === null) {
+                                this.throwUnexpectedToken(Messages.InvalidHexEscapeSequence);
+                            }
+                            cooked += unescaped;
+                            break;
                         case 'b':
                             cooked += '\b';
                             break;
@@ -1008,10 +1071,8 @@ export class Scanner {
 
         return {
             type: Token.Template,
-            value: {
-                cooked: cooked,
-                raw: this.source.slice(start + 1, this.index - rawOffset)
-            },
+            value: this.source.slice(start + 1, this.index - rawOffset),
+            cooked: cooked,
             head: head,
             tail: tail,
             lineNumber: this.lineNumber,
@@ -1019,11 +1080,11 @@ export class Scanner {
             start: start,
             end: this.index
         };
-    };
+    }
 
-    // ECMA-262 11.8.5 Regular Expression Literals
+    // https://tc39.github.io/ecma262/#sec-literals-regular-expression-literals
 
-    testRegExp(pattern: string, flags: string) {
+    private testRegExp(pattern: string, flags: string): RegExp | null {
         // The BMP character to use as a replacement for astral symbols when
         // translating an ES6 "u"-flagged pattern to an ES5-compatible
         // approximation.
@@ -1032,7 +1093,7 @@ export class Scanner {
         // pattern that would not be detected by this substitution.
         const astralSubstitute = '\uFFFF';
         let tmp = pattern;
-        let self = this;
+        const self = this;
 
         if (flags.indexOf('u') >= 0) {
             tmp = tmp
@@ -1040,7 +1101,7 @@ export class Scanner {
                 // BMP character or a constant ASCII code point in the case of
                 // astral symbols. (See the above note on `astralSubstitute`
                 // for more information.)
-                .replace(/\\u\{([0-9a-fA-F]+)\}|\\u([a-fA-F0-9]{4})/g, function($0, $1, $2) {
+                .replace(/\\u\{([0-9a-fA-F]+)\}|\\u([a-fA-F0-9]{4})/g, ($0, $1, $2) => {
                     const codePoint = parseInt($1 || $2, 16);
                     if (codePoint > 0x10FFFF) {
                         self.throwUnexpectedToken(Messages.InvalidRegExp);
@@ -1075,9 +1136,9 @@ export class Scanner {
             /* istanbul ignore next */
             return null;
         }
-    };
+    }
 
-    scanRegExpBody() {
+    private scanRegExpBody(): string {
         let ch = this.source[this.index];
         assert(ch === '/', 'Regular expression literal must start with a slash');
 
@@ -1090,7 +1151,7 @@ export class Scanner {
             str += ch;
             if (ch === '\\') {
                 ch = this.source[this.index++];
-                // ECMA-262 7.8.5
+                // https://tc39.github.io/ecma262/#sec-literals-regular-expression-literals
                 if (Character.isLineTerminator(ch.charCodeAt(0))) {
                     this.throwUnexpectedToken(Messages.UnterminatedRegExp);
                 }
@@ -1116,14 +1177,10 @@ export class Scanner {
         }
 
         // Exclude leading and trailing slash.
-        const body = str.substr(1, str.length - 2);
-        return {
-            value: body,
-            literal: str
-        };
-    };
+        return str.substr(1, str.length - 2);
+    }
 
-    scanRegExpFlags() {
+    private scanRegExpFlags(): string {
         let str = '';
         let flags = '';
         while (!this.eof()) {
@@ -1138,9 +1195,9 @@ export class Scanner {
                 if (ch === 'u') {
                     ++this.index;
                     let restore = this.index;
-                    ch = this.scanHexEscape('u');
-                    if (ch) {
-                        flags += ch;
+                    const char = this.scanHexEscape('u');
+                    if (char !== null) {
+                        flags += char;
                         for (str += '\\u'; restore < this.index; ++restore) {
                             str += this.source[restore];
                         }
@@ -1160,38 +1217,34 @@ export class Scanner {
             }
         }
 
-        return {
-            value: flags,
-            literal: str
-        };
-    };
+        return flags;
+    }
 
-    scanRegExp() {
+    public scanRegExp(): RawToken {
         const start = this.index;
 
-        const body = this.scanRegExpBody();
+        const pattern = this.scanRegExpBody();
         const flags = this.scanRegExpFlags();
-        const value = this.testRegExp(body.value, flags.value);
+        const value = this.testRegExp(pattern, flags);
 
         return {
             type: Token.RegularExpression,
-            value: value,
-            literal: body.literal + flags.literal,
-            regex: {
-                pattern: body.value,
-                flags: flags.value
-            },
+            value: '',
+            pattern: pattern,
+            flags: flags,
+            regex: value,
             lineNumber: this.lineNumber,
             lineStart: this.lineStart,
             start: start,
             end: this.index
         };
-    };
+    }
 
-    lex() {
+    public lex(): RawToken {
         if (this.eof()) {
             return {
                 type: Token.EOF,
+                value: '',
                 lineNumber: this.lineNumber,
                 lineStart: this.lineStart,
                 start: this.index,
@@ -1242,6 +1295,6 @@ export class Scanner {
         }
 
         return this.scanPunctuator();
-    };
+    }
 
 }
