@@ -108,7 +108,8 @@ if (typeof child_process.execSync !== 'function') {
 }
 
 test_downstream({
-    'escope': 'https://github.com/estools/escope.git',
+    'envify': 'https://github.com/hughsk/envify.git',
+    // 'escope': 'https://github.com/estools/escope.git',
     'esmangle': 'https://github.com/estools/esmangle.git',
     'escomplex-js': 'https://github.com/philbooth/escomplex-js.git',
     // 'js2coffee': 'https://github.com/js2coffee/js2coffee.git',
@@ -116,7 +117,7 @@ test_downstream({
     'jsfmt': 'https://github.com/rdio/jsfmt.git',
     'assetgraph': 'https://github.com/assetgraph/assetgraph.git',
     'recast': 'https://github.com/benjamn/recast.git',
-    'rocambole': 'https://github.com/millermedeiros/rocambole.git',
-    'documentjs': 'https://github.com/bitovi/documentjs.git'
+    'rocambole': 'https://github.com/millermedeiros/rocambole.git'
+    // 'documentjs': 'https://github.com/bitovi/documentjs.git'
     // 'istanbul': 'https://github.com/gotwarlost/istanbul.git'
 });
