@@ -41,9 +41,11 @@ module.exports = function (config) {
         plugins: [
             'karma-mocha',
             'karma-chrome-launcher',
+            'karma-edge-launcher',
             'karma-firefox-launcher',
             'karma-ie-launcher',
             'karma-safari-launcher',
+            'karma-safaritechpreview-launcher',
             'karma-detect-browsers'
         ]
 
